@@ -6,6 +6,8 @@ importing sub-modules.
 """
 
 import contextlib
+from datetime import datetime
+from os.path import getctime, getmtime
 from pathlib import Path
 from urllib.parse import urljoin
 
@@ -13,7 +15,7 @@ import cssselect2
 import html5lib
 import tinycss2
 
-VERSION = __version__ = '60.1'
+VERSION = __version__ = '61.0'
 
 #: Default values for command-line and Python API options. See
 #: :func:`__main__.main` to learn more about specific options for
@@ -133,7 +135,8 @@ class HTML:
 
     :param str encoding:
         Force the source character encoding.
-    :param str base_url:
+    :type base_url: str or pathlib.Path
+    :param base_url:
         The base used to resolve relative URLs (e.g. in
         ``<img src="../foo.png">``). If not provided, try to use the input
         filename, URL, or ``name`` attribute of
@@ -156,16 +159,20 @@ class HTML:
             'Step 1 - Fetching and parsing HTML - %s',
             guess or filename or url or
             getattr(file_obj, 'name', 'HTML string'))
+        if isinstance(base_url, Path):
+            base_url = str(base_url)
         result = _select_source(
             guess, filename, url, file_obj, string, base_url, url_fetcher)
         with result as (source_type, source, base_url, protocol_encoding):
             if isinstance(source, str):
                 result = html5lib.parse(source, namespaceHTMLElements=False)
             else:
-                result = html5lib.parse(
-                    source, override_encoding=encoding,
-                    transport_encoding=protocol_encoding,
-                    namespaceHTMLElements=False)
+                kwargs = {'namespaceHTMLElements': False}
+                if protocol_encoding is not None:
+                    kwargs['transport_encoding'] = protocol_encoding
+                if encoding is not None:
+                    kwargs['override_encoding'] = encoding
+                result = html5lib.parse(source, **kwargs)
         self.base_url = _find_base_url(result, base_url)
         self.url_fetcher = url_fetcher
         self.media_type = media_type
@@ -281,12 +288,12 @@ class CSS:
             base_url=base_url, url_fetcher=url_fetcher,
             check_css_mime_type=_check_mime_type)
         with result as (source_type, source, base_url, protocol_encoding):
-            if source_type == 'string' and not isinstance(source, bytes):
+            if source_type == 'file_obj':
+                source = source.read()
+            if isinstance(source, str):
                 # unicode, no encoding
                 stylesheet = tinycss2.parse_stylesheet(source)
             else:
-                if source_type == 'file_obj':
-                    source = source.read()
                 stylesheet, encoding = tinycss2.parse_stylesheet_bytes(
                     source, environment_encoding=encoding,
                     protocol_encoding=protocol_encoding)
@@ -303,21 +310,46 @@ class Attachment:
 
     An instance is created in the same way as :class:`HTML`, except that the
     HTML specific arguments (``encoding`` and ``media_type``) are not
-    supported. An optional description can be provided with the ``description``
-    argument.
+    supported.
 
-    :param description:
+    :param str description:
         A description of the attachment to be included in the PDF document.
         May be :obj:`None`.
+    :type created: :obj:`datetime.datetime`
+    :param created:
+        Creation date and time. Default is current date and time.
+    :type modified: :obj:`datetime.datetime`
+    :param modified:
+        Modification date and time. Default is current date and time.
+    :param str relationship:
+        A string that represents the relationship between the attachment and
+        the PDF it is embedded in. Default is 'Unspecified', other common
+        values are defined in ISO-32000-2:2020, 7.11.3.
 
     """
     def __init__(self, guess=None, filename=None, url=None, file_obj=None,
                  string=None, base_url=None, url_fetcher=default_url_fetcher,
-                 description=None):
+                 description=None, created=None, modified=None,
+                 relationship='Unspecified'):
         self.source = _select_source(
             guess, filename, url, file_obj, string, base_url=base_url,
             url_fetcher=url_fetcher)
         self.description = description
+        self.relationship = relationship
+        self.md5 = None
+
+        if created is None:
+            if filename:
+                created = datetime.fromtimestamp(getctime(filename))
+            else:
+                created = datetime.now()
+        if modified is None:
+            if filename:
+                modified = datetime.fromtimestamp(getmtime(filename))
+            else:
+                modified = datetime.now()
+        self.created = created
+        self.modified = modified
 
 
 @contextlib.contextmanager
