@@ -17,14 +17,19 @@ from logging import DEBUG, WARNING
 
 import cssselect2
 import tinycss2
+import tinycss2.ast
 import tinycss2.nth
 
 from .. import CSS
 from ..logger import LOGGER, PROGRESS_LOGGER
 from ..urls import URLFetchingError, get_url_attribute, url_join
-from . import computed_values, counters, media_queries
-from .properties import INHERITED, INITIAL_NOT_COMPUTED, INITIAL_VALUES
-from .utils import get_url, remove_whitespace
+from . import counters, media_queries
+from .computed_values import COMPUTER_FUNCTIONS
+from .properties import (
+    INHERITED, INITIAL_NOT_COMPUTED, INITIAL_VALUES, ZERO_PIXELS)
+from .utils import (
+    InvalidValues, Pending, check_var_function, get_url, parse_function,
+    remove_whitespace)
 from .validation import preprocess_declarations
 from .validation.descriptors import preprocess_descriptors
 
@@ -124,13 +129,13 @@ class StyleFor:
                     style['border_collapse'] == 'collapse'):
                 # Padding do not apply
                 for side in ('top', 'bottom', 'left', 'right'):
-                    style[f'padding_{side}'] = computed_values.ZERO_PIXELS
+                    style[f'padding_{side}'] = ZERO_PIXELS
             if (len(style['display']) == 1 and
                     style['display'][0].startswith('table-') and
                     style['display'][0] != 'table-caption'):
                 # Margins do not apply
                 for side in ('top', 'bottom', 'left', 'right'):
-                    style[f'margin_{side}'] = computed_values.ZERO_PIXELS
+                    style[f'margin_{side}'] = ZERO_PIXELS
 
         return style
 
@@ -163,18 +168,6 @@ class StyleFor:
             element, cascaded, parent_style, pseudo_type, root_style, base_url,
             target_collector)
 
-        # The style of marker is deleted when display is different from
-        # list-item.
-        if pseudo_type is None:
-            for pseudo in (None, 'before', 'after'):
-                pseudo_style = cascaded_styles.get((element, pseudo), {})
-                if 'display' in pseudo_style:
-                    if 'list-item' in pseudo_style['display'][0]:
-                        break
-            else:
-                if (element, 'marker') in cascaded_styles:
-                    del cascaded_styles[element, 'marker']
-
     def add_page_declarations(self, page_type):
         for sheet, origin, sheet_specificity in self._sheets:
             for _rule, selector_list, declarations in sheet.page_rules:
@@ -600,6 +593,36 @@ def declaration_precedence(origin, importance):
         return 5
 
 
+def resolve_var(computed, token, parent_style):
+    """Return token with resolved CSS variables."""
+    if not check_var_function(token):
+        return
+
+    if token.lower_name != 'var':
+        arguments = []
+        for i, argument in enumerate(token.arguments):
+            if argument.type == 'function' and argument.lower_name == 'var':
+                arguments.extend(resolve_var(computed, argument, parent_style))
+            else:
+                arguments.append(argument)
+        token = tinycss2.ast.FunctionBlock(
+            token.source_line, token.source_column, token.name, arguments)
+        return resolve_var(computed, token, parent_style) or (token,)
+
+    args = parse_function(token)[1]
+    variable_name = args.pop(0).value.replace('-', '_')  # first arg is name
+    default = args  # next args are default value
+    computed_value = []
+    for value in computed[variable_name]:
+        resolved = resolve_var(computed, value, parent_style)
+        computed_value.extend((value,) if resolved is None else resolved)
+    if len(computed_value) == 1:
+        token, = computed_value
+        if token.type == 'ident' and token.value == 'initial':
+            return default
+    return computed_value
+
+
 class AnonymousStyle(dict):
     """Computed style used for anonymous boxes."""
     def __init__(self, parent_style):
@@ -667,75 +690,91 @@ class ComputedStyle(dict):
 
     def __missing__(self, key):
         if key == 'float':
-            # Set specified value for position, needed for computed value
+            # Set specified value for position, needed for computed value.
             self['position']
         elif key == 'display':
-            # Set specified value for float, needed for computed value
+            # Set specified value for float, needed for computed value.
             self['float']
 
+        parent_style = self.parent_style
+
         if key in self.cascaded:
-            value = keyword = self.cascaded[key][0]
+            # Property defined in cascaded properties.
+            value = self.cascaded[key][0]
+            pending = isinstance(value, Pending)
         else:
+            # Property not defined in cascaded properties, define as inherited
+            # or initial value.
             if key in INHERITED or key[:2] == '__':
-                keyword = 'inherit'
+                value = 'inherit'
             else:
-                keyword = 'initial'
+                value = 'initial'
+            pending = False
 
-        if keyword == 'inherit' and self.parent_style is None:
+        if value == 'inherit' and parent_style is None:
             # On the root element, 'inherit' from initial values
-            keyword = 'initial'
+            value = 'initial'
 
-        if keyword == 'initial':
-            value = None if key[:2] == '__' else INITIAL_VALUES[key]
+        if pending:
+            # Property with pending values, validate them.
+            solved_tokens = []
+            for token in value.tokens:
+                tokens = resolve_var(self, token, parent_style)
+                if tokens is None:
+                    solved_tokens.append(token)
+                else:
+                    solved_tokens.extend(tokens)
+            original_key = key.replace('_', '-')
+            try:
+                value = value.solve(solved_tokens, original_key)
+            except InvalidValues:
+                if key in INHERITED:
+                    # Values in parent_style are already computed.
+                    self[key] = value = parent_style[key]
+                else:
+                    value = INITIAL_VALUES[key]
+                    if key not in INITIAL_NOT_COMPUTED:
+                        # The value is the same as when computed.
+                        self[key] = value
+
+        if value == 'initial':
+            value = [] if key[:2] == '__' else INITIAL_VALUES[key]
             if key not in INITIAL_NOT_COMPUTED:
-                # The value is the same as when computed
+                # The value is the same as when computed.
                 self[key] = value
-        elif keyword == 'inherit':
+        elif value == 'inherit':
             # Values in parent_style are already computed.
-            self[key] = value = self.parent_style[key]
+            self[key] = value = parent_style[key]
 
-        if key[:16] == 'text_decoration_' and self.parent_style:
+        if key[:16] == 'text_decoration_' and parent_style is not None:
+            # Text decorations are not inherited but propagated. See
+            # https://www.w3.org/TR/css-text-decor-3/#line-decoration.
             value = text_decoration(
-                key, value, self.parent_style[key], key in self.cascaded)
+                key, value, parent_style[key], key in self.cascaded)
             if key in self:
                 del self[key]
         elif key == 'page' and value == 'auto':
-            # The page property does not inherit. However, if the page
-            # value on an element is auto, then its used value is the value
-            # specified on its nearest ancestor with a non-auto value. When
-            # specified on the root element, the used value for auto is the
-            # empty string.
-            value = (
-                '' if self.parent_style is None else self.parent_style['page'])
+            # The page property does not inherit. However, if the page value on
+            # an element is auto, then its used value is the value specified on
+            # its nearest ancestor with a non-auto value. When specified on the
+            # root element, the used value for auto is the empty string. See
+            # https://www.w3.org/TR/css-page-3/#using-named-pages.
+            value = '' if parent_style is None else parent_style['page']
             if key in self:
                 del self[key]
         elif key in ('position', 'float', 'display'):
+            # Save specified values to define computed values for these
+            # specific properties. See
+            # https://www.w3.org/TR/CSS21/visuren.html#dis-pos-flo.
             self.specified[key] = value
 
         if key in self:
+            # Value already computed and saved: return.
             return self[key]
 
-        function = computed_values.COMPUTER_FUNCTIONS.get(key)
-        already_computed_value = False
-
-        if value:
-            converted_to_list = False
-
-            if not isinstance(value, list):
-                converted_to_list = True
-                value = [value]
-
-            for i, v in enumerate(value):
-                value[i], already_computed_value = (
-                    computed_values.compute_variable(
-                        v, key, self, self.base_url, self.parent_style))
-
-            if converted_to_list:
-                value, = value
-
-        if function is not None and not already_computed_value:
-            value = function(self, key, value)
-        # else: same as specified
+        if key in COMPUTER_FUNCTIONS:
+            # Value not computed yet: compute.
+            value = COMPUTER_FUNCTIONS[key](self, key, value)
 
         self[key] = value
         return value
