@@ -9,7 +9,8 @@ from cssselect2 import ElementWrapper
 
 from ..urls import get_url_attribute
 from .bounding_box import (
-    bounding_box, extend_bounding_box, is_valid_bounding_box)
+    EMPTY_BOUNDING_BOX, bounding_box, extend_bounding_box,
+    is_valid_bounding_box)
 from .css import parse_declarations, parse_stylesheets
 from .defs import (
     apply_filters, clip_path, draw_gradient_or_pattern, paint_mask, use)
@@ -18,7 +19,8 @@ from .path import path
 from .shapes import circle, ellipse, line, polygon, polyline, rect
 from .text import text
 from .utils import (
-    PointError, color, normalize, parse_url, preserve_ratio, size, transform)
+    PointError, alpha_value, color, normalize, parse_url, preserve_ratio, size,
+    transform)
 
 TAGS = {
     'a': text,
@@ -119,6 +121,16 @@ class Node:
         """Text after the XML node."""
         return self._etree_node.tail
 
+    @property
+    def display(self):
+        """Whether node should be displayed."""
+        return self.get('display') != 'none'
+
+    @property
+    def visible(self):
+        """Whether node is visible."""
+        return self.display and self.get('visibility') != 'hidden'
+
     def __iter__(self):
         """Yield node children, handling cascade."""
         for wrapper in self._wrapper:
@@ -387,6 +399,8 @@ class SVG:
         # Update font size
         font_size = size(node.get('font-size', '1em'), font_size, font_size)
 
+        original_streams = []
+
         if fill_stroke:
             self.stream.push_state()
 
@@ -399,9 +413,9 @@ class SVG:
         self.transform(node.get('transform'), font_size)
 
         # Create substream for opacity
-        opacity = float(node.get('opacity', 1))
+        opacity = alpha_value(node.get('opacity', 1))
         if fill_stroke and 0 <= opacity < 1:
-            original_stream = self.stream
+            original_streams.append(self.stream)
             box = self.calculate_bounding_box(node, font_size)
             if not is_valid_bounding_box(box):
                 box = (0, 0, self.inner_width, self.inner_height)
@@ -430,26 +444,27 @@ class SVG:
             if new_ctm.determinant:
                 self.stream.transform(*(old_ctm @ new_ctm.invert).values)
 
-        # Manage display and visibility
-        display = node.get('display') != 'none'
-        visible = display and (node.get('visibility') != 'hidden')
-
         # Handle text anchor
         text_anchor = node.get('text-anchor')
-        if node.tag == 'text' and text_anchor in ('middle', 'end'):
+        if TAGS.get(node.tag) == text and text_anchor in ('middle', 'end'):
             group = self.stream.add_group(0, 0, 0, 0)  # BBox set after drawing
-            original_stream, self.stream = self.stream, group
+            original_streams.append(self.stream)
+            self.stream = group
+
+        # Set text bounding box
+        if node.display and TAGS.get(node.tag) == text:
+            node.text_bounding_box = EMPTY_BOUNDING_BOX
 
         # Draw node
-        if visible and node.tag in TAGS:
+        if node.visible and node.tag in TAGS:
             with suppress(PointError):
                 TAGS[node.tag](self, node, font_size)
 
         # Draw node children
-        if display and node.tag not in DEF_TYPES:
+        if node.display and node.tag not in DEF_TYPES:
             for child in node:
                 self.draw_node(child, font_size, fill_stroke)
-                if node.tag in ('text', 'tspan'):
+                if TAGS.get(node.tag) == text and child.visible:
                     if not is_valid_bounding_box(child.text_bounding_box):
                         continue
                     x1, y1 = child.text_bounding_box[:2]
@@ -459,9 +474,9 @@ class SVG:
                         node.text_bounding_box, ((x1, y1), (x2, y2)))
 
         # Handle text anchor
-        if node.tag == 'text' and text_anchor in ('middle', 'end'):
+        if TAGS.get(node.tag) == text and text_anchor in ('middle', 'end'):
             group_id = self.stream.id
-            self.stream = original_stream
+            self.stream = original_streams.pop()
             self.stream.push_state()
             if is_valid_bounding_box(node.text_bounding_box):
                 x, y, width, height = node.text_bounding_box
@@ -486,7 +501,7 @@ class SVG:
         # Apply opacity stream and restore original stream
         if fill_stroke and 0 <= opacity < 1:
             group_id = self.stream.id
-            self.stream = original_stream
+            self.stream = original_streams.pop()
             self.stream.set_alpha(opacity, stroke=True, fill=True)
             self.stream.draw_x_object(group_id)
 
@@ -648,7 +663,7 @@ class SVG:
 
         # Get fill data
         fill_source, fill_color = self.get_paint(node.get('fill', 'black'))
-        fill_opacity = float(node.get('fill-opacity', 1))
+        fill_opacity = alpha_value(node.get('fill-opacity', 1))
         fill_drawn = draw_gradient_or_pattern(
             self, node, fill_source, font_size, fill_opacity, stroke=False)
         if fill_color and not fill_drawn:
@@ -659,7 +674,7 @@ class SVG:
 
         # Get stroke data
         stroke_source, stroke_color = self.get_paint(node.get('stroke'))
-        stroke_opacity = float(node.get('stroke-opacity', 1))
+        stroke_opacity = alpha_value(node.get('stroke-opacity', 1))
         stroke_drawn = draw_gradient_or_pattern(
             self, node, stroke_source, font_size, stroke_opacity, stroke=True)
         if stroke_color and not stroke_drawn:
